#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "lib.h"
#include "iwram.h"
#include "irq.h"
#include "syscall.h"

#include "resource.h"
#include "resource\BGFont.h"
//#include "resource\BitMapFont.h"

//#include <string.h>
//#include <stdlib.h>

//---------------------------------------------------------------------------
#define BULLET_MAX           100
#define BULLET_SPEED         1
#define BULLET_STATE_RELEASE 1
#define BULLET_STATE_FREE    2
#define BULLET_MAKE_TIMING   32

#define UNIT_VECT_UP    0x01
#define UNIT_VECT_DOWN  0x02
#define UNIT_VECT_RIGHT 0x04
#define UNIT_VECT_LEFT  0x08
#define UNIT_VECT_NONE  0x00
//---------------------------------------------------------------------------
typedef struct {
	u8  state;		//
	u8  angle;		//px
	f32 speed;		//Xs[h
	f32 x;
	f32 y;
	f32 txv;		//ړ
	f32 tyv;
	u8  cx;
	u8  cy;
	u16 SprNo;
	u16 SprTile;
	u16 SprSize;
	u16 SprForm;

	void (*pAction)(u8);
	u8   data1;
	u8   data2;
	u8   data3;
}ST_BULLET;

ST_BULLET BULLET[BULLET_MAX];

typedef struct {
	u8  x;
	u8  y;
	u8  cx;
	u8  cy;
	u8  vect;		//ړ
	u16 SprNo;
	u16 SprTile;
	u16 SprSize;
	u16 SprForm;
}ST_UNIT;

ST_UNIT   PLAYER;

//---------------------------------------------------------------------------
u8 BulletCnt;
u8 BulletMode;
u8 BulletMakeTiming;
u8 BulletMakeTiming2;

u8 BG2Chr;
u8 BG2Map;
u8 BG2Col;
//---------------------------------------------------------------------------
void Init();
void InitSprite();

void Player();
void Bullet();
void HitProcess();
void MoveProcess();
void BulletProcess();

u8   BulletRegist(s16 x, s16 y, u8 angle, u8 speed, void* pAction);
void BulletRegistWay(s16 x, s16 y, u16 angleAdd, u8 cnt);
void BulletRegistDirect(s16 x, s16 y);
void BulletRegistCircle(s16 x, s16 y, u16 angle);
void BulletRegistSplit(s16 x, s16 y);
void BulletRegistHoming(s16 x, s16 y);
void BulletRegistAccelerate(s16 x, s16 y);
void BulletRegistLoop(s16 x, s16 y, u16 angle);

void BulletActionSplit(u8);
void BulletActionHoming(u8);
void BulletActionAccelerate(u8);
void BulletActionLoop(u8);

u8   BulletCalcAngle(s16 x, s16 y);

void BulletDestory(u8 i);

u8   ChkMoveUnit(ST_UNIT* p);
void MoveUnit(ST_UNIT* p);
void DrawUnit(ST_UNIT* p);

u8   ChkMoveBullet(ST_BULLET* p);
void MoveBullet(ST_BULLET* p);
void DrawBullet(ST_BULLET* p);

void DrawBulletCnt();
//---------------------------------------------------------------------------
int main(void)
{
	Init();
	InitSprite();
	SetMode(MODE_0 | BG2_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	for(;;)
	{
		WaitForVsync();

		MoveProcess();
		BulletProcess();
		HitProcess();

		Player();
		Bullet();

		DrawBulletCnt();
	}
	return 0;
}
//---------------------------------------------------------------------------
void Init()
{
	BulletCnt = 0;
	BulletMode = 0;
	BulletMakeTiming  = 0;
	BulletMakeTiming2 = 0;

	//BG
	BG2Chr = 1;
	BG2Map = 0;
	BG2Col = 0;

	REG_BG2CNT = (BG_SIZEA_256_256    | BG_COLOR_16        |
				  BG_CHARBASE(BG2Chr) | BG_MAPBASE(BG2Map) | 0);

	BGSetPal16  (BG2Col, (u16*)BGFont_palette);
	BGSetChrData(BG2Chr, (u16*)BGFont_data, BGFont_DATA_SIZE);

	//Sprite
	SpriteInit();
	SpriteSetPal16((u16*)&game_spr_pal_bin);
	SpriteSetData ((u16*)&game_spr_img_bin, 2048);

	//Key
	KeyInit();
}
//---------------------------------------------------------------------------
void InitSprite()
{
	//PLAYER
	PLAYER.cx      = 8;
	PLAYER.cy      = 8;
	PLAYER.x       = SCREEN_CX/2 - PLAYER.cx/2;
	PLAYER.y       = SCREEN_CY/2 - PLAYER.cy/2;
	PLAYER.vect    = UNIT_VECT_NONE;
	PLAYER.SprNo   = 0;
	PLAYER.SprTile = 0;
	PLAYER.SprSize = SP_SIZE_8;
	PLAYER.SprForm = SP_SQUARE;

	SpriteSetSize(PLAYER.SprNo, PLAYER.SprSize, PLAYER.SprForm, SP_COLOR_16);
	SpriteSetChr (PLAYER.SprNo, PLAYER.SprTile);
	SpriteMove   (PLAYER.SprNo, PLAYER.x, PLAYER.y);

	//BULLET
	u8 i;
	for(i=0; i<BULLET_MAX; i++)
	{
		BULLET[i].state   = BULLET_STATE_FREE;
		BULLET[i].x       = 240;
		BULLET[i].y       = 160;
		BULLET[i].cx      = 8;
		BULLET[i].cy      = 8;
		BULLET[i].SprNo   = i+1;
		BULLET[i].SprTile = 1;
		BULLET[i].SprSize = SP_SIZE_8;
		BULLET[i].SprForm = SP_SQUARE;

		SpriteSetSize(BULLET[i].SprNo, BULLET[i].SprSize, BULLET[i].SprForm, SP_COLOR_16);
		SpriteSetChr (BULLET[i].SprNo, BULLET[i].SprTile);
		SpriteMove   (BULLET[i].SprNo, BULLET[i].x, BULLET[i].y);
	}
}
//---------------------------------------------------------------------------
//߂l͍쐬łǂ
u8 BulletRegist(s16 x, s16 y, u8 angle, u8 speed, void* pAction)
{
	u8 i;
	for(i=0; i<BULLET_MAX; i++)
	{
		if(BULLET[i].state == BULLET_STATE_FREE)
		{
			BULLET[i].state   = BULLET_STATE_RELEASE;
			BULLET[i].x       = x;
			BULLET[i].y       = y;
			BULLET[i].txv     = speed * _Cos(angle);
			BULLET[i].tyv     = speed * _Sin(angle);
			BULLET[i].angle   = angle;
			BULLET[i].speed   = speed;

			BULLET[i].pAction = pAction;
			BULLET[i].data1   = 0;
			BULLET[i].data2   = 0;
			BULLET[i].data3   = 0;

			BulletCnt++;
			return TRUE;
		}
	}
	return FALSE;
}
//---------------------------------------------------------------------------
void BulletRegistDirect(s16 x, s16 y)
{
	u8 angle = BulletCalcAngle(x, y);
	BulletRegist(x, y, angle, BULLET_SPEED, NULL);
}
//---------------------------------------------------------------------------
void BulletRegistSplit(s16 x, s16 y)
{
	u8 angle = BulletCalcAngle(x, y);
	BulletRegist(x, y, angle, BULLET_SPEED, &BulletActionSplit);
}
//---------------------------------------------------------------------------
void BulletRegistAccelerate(s16 x, s16 y)
{
	u8 angle = BulletCalcAngle(x, y);
	BulletRegist(x, y, angle, BULLET_SPEED, &BulletActionAccelerate);
}
//---------------------------------------------------------------------------
void BulletRegistHoming(s16 x, s16 y)
{
	u8 angle = BulletCalcAngle(x, y);
	BulletRegist(x, y, angle, BULLET_SPEED, &BulletActionHoming);
}
//---------------------------------------------------------------------------
void BulletRegistWay(s16 x, s16 y, u16 angleAdd, u8 num)
{
	BulletRegistDirect(x, y);

	u8  angle = BulletCalcAngle(x, y);
	u16 i;
	for(i=1; i<num+1; i++)
	{
		BulletRegist(x, y, angle + angleAdd * i, BULLET_SPEED, NULL);
		BulletRegist(x, y, angle - angleAdd * i, BULLET_SPEED, NULL);
	}
}
//---------------------------------------------------------------------------
void BulletRegistCircle(s16 x, s16 y, u16 angle)
{
	u16 i;
	for(i=0; i<256; i+=angle)
	{
		BulletRegist(x, y, i, BULLET_SPEED, NULL);
	}
}
//---------------------------------------------------------------------------
void BulletRegistLoop(s16 x, s16 y, u16 angle)
{
	u16 i;
	for(i=0; i<256; i+=angle)
	{
		BulletRegist(x, y, i, BULLET_SPEED, &BulletActionLoop);
	}
}
//---------------------------------------------------------------------------
void BulletActionSplit(u8 i)
{
	if(BULLET[i].data1++ > 48)
	{
		BulletRegistWay(BULLET[i].x, BULLET[i].y, 10, 1);
		BulletDestory(i);
	}
}
//---------------------------------------------------------------------------
void BulletActionHoming(u8 i)
{
	//data1 vZ^C~O
	
	if(BULLET[i].data1++ > 5)
	{
		BULLET[i].data1 = 0;

		s16 OldAngle = (s16)BULLET[i].angle;
		s16 NewAngle = (s16)BulletCalcAngle(BULLET[i].x, BULLET[i].y);
		s16 ret      = NewAngle - OldAngle;

		if( ret<=-128 || (ret>0 && ret<=128) )
		{
			BULLET[i].angle += 6;
		}
		else if(ret>128 || (ret<0 && ret>-128) )
		{
			BULLET[i].angle -= 6;
		}
		else
		{
			BULLET[i].angle = NewAngle;
		}

		BULLET[i].txv = BULLET[i].speed * _Cos(BULLET[i].angle);
		BULLET[i].tyv = BULLET[i].speed * _Sin(BULLET[i].angle);
	}
}
//---------------------------------------------------------------------------
void BulletActionAccelerate(u8 i)
{
	BULLET[i].txv *= 1.03;
	BULLET[i].tyv *= 1.03;
}
//---------------------------------------------------------------------------
void BulletActionLoop(u8 i)
{
	if(BULLET[i].data1++ > 1)
	{
		BULLET[i].data1 = 0;

		BULLET[i].angle-=3;
		BULLET[i].speed*=1.02;
		BULLET[i].txv = BULLET[i].speed * _Cos(BULLET[i].angle);
		BULLET[i].tyv = BULLET[i].speed * _Sin(BULLET[i].angle);
	}
}
//---------------------------------------------------------------------------
u8 BulletCalcAngle(s16 x, s16 y)
{
	return GetAngle(PLAYER.x - x, PLAYER.y - y);
}
//---------------------------------------------------------------------------
void Player()
{
	u8 Key = KeyInput();

	PLAYER.vect = UNIT_VECT_NONE;

	if(Key & KEY_UP   ) PLAYER.vect |= UNIT_VECT_UP;
	if(Key & KEY_DOWN ) PLAYER.vect |= UNIT_VECT_DOWN;
	if(Key & KEY_LEFT ) PLAYER.vect |= UNIT_VECT_LEFT;
	if(Key & KEY_RIGHT) PLAYER.vect |= UNIT_VECT_RIGHT;

	if( ChkMoveUnit(&PLAYER) != TRUE )
	{
		PLAYER.vect = UNIT_VECT_NONE;
	}
}
//---------------------------------------------------------------------------
void Bullet()
{
	BulletMakeTiming++;
	BulletMakeTiming &= BULLET_MAKE_TIMING - 1;

	if(BulletMakeTiming != 0)
	{
		return;
	}

	if(BulletMakeTiming2++ > 8)
	{
		BulletMakeTiming2 = 0;
		BulletMode++;

		if(BulletMode>6) BulletMode=0;
	}

	switch(BulletMode)
	{
	case 0: BulletRegistDirect(10,10);            break;
	case 1: BulletRegistSplit(10, 160-10);        break;
	case 2: BulletRegistWay(240-10, 10, 5, 1);    break;
	case 3: BulletRegistHoming(240-10, 160-10);   break;
	case 4: BulletRegistAccelerate(10, 10);       break;
	case 5: BulletRegistCircle(240/2, 160/2, 16); break;
	case 6: BulletRegistLoop(240/2, 160/2, 16);   break;
	default: break;
	}

}
//---------------------------------------------------------------------------
void BulletProcess()
{
	u8 i;
	for(i=0; i<BULLET_MAX; i++)
	{
		if(BULLET[i].state == BULLET_STATE_RELEASE)
		{
			if( ChkMoveBullet(&BULLET[i]) != TRUE )
			{
				BulletDestory(i);
			}
			else if(BULLET[i].pAction != NULL)
			{
				BULLET[i].pAction(i);
			}
		}
	}
}
//---------------------------------------------------------------------------
void BulletDestory(u8 i)
{
	BULLET[i].state = BULLET_STATE_FREE;
	BULLET[i].x     = 240;
	BULLET[i].y     = 160;
	DrawBullet(&BULLET[i]);

	BulletCnt--;
}
//---------------------------------------------------------------------------
//ړ͈ړ\`FbN
u8 ChkMoveUnit(ST_UNIT* p)
{
	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(p->vect & UNIT_VECT_UP   ) ty1--;
	if(p->vect & UNIT_VECT_DOWN ) ty1++;
	if(p->vect & UNIT_VECT_LEFT ) tx1--;
	if(p->vect & UNIT_VECT_RIGHT) tx1++;

	s16 tx2 = tx1 + p->cx -1;
	s16 ty2 = ty1 + p->cy -1;

	if(tx1<   0 || ty1<   0) return FALSE;
	if(tx2>=240 || ty2>=160) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
u8 ChkMoveBullet(ST_BULLET* p)
{
	s16 tx1 = p->x + p->txv;
	s16 ty1 = p->y + p->tyv;
	s16 tx2 = tx1  + p->cx -1;
	s16 ty2 = ty1  + p->cy -1;

	if(tx1<   0 || ty1<   0) return FALSE;
	if(tx2>=240 || ty2>=160) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
void MoveProcess()
{
	//PLAYER
	if(PLAYER.vect != UNIT_VECT_NONE)
	{
		MoveUnit(&PLAYER);
	}

	//BULLET
	u8 i;
	for(i=0; i<BULLET_MAX; i++)
	{
		if(BULLET[i].state == BULLET_STATE_RELEASE)
		{
			MoveBullet(&BULLET[i]);
		}
	}
}
//---------------------------------------------------------------------------
void MoveUnit(ST_UNIT* p)
{
	if(p->vect & UNIT_VECT_UP   ) p->y--;
	if(p->vect & UNIT_VECT_DOWN ) p->y++;
	if(p->vect & UNIT_VECT_LEFT ) p->x--;
	if(p->vect & UNIT_VECT_RIGHT) p->x++;

	DrawUnit(p);
}
//---------------------------------------------------------------------------
void MoveBullet(ST_BULLET* p)
{
	p->x += p->txv;
	p->y += p->tyv;

	DrawBullet(p);
}
//---------------------------------------------------------------------------
void DrawUnit(ST_UNIT* p)
{
	SpriteMove(p->SprNo, p->x, p->y);
}
//---------------------------------------------------------------------------
void DrawBullet(ST_BULLET* p)
{
	SpriteMove(p->SprNo, p->x, p->y);
}
//---------------------------------------------------------------------------
void HitProcess()
{
}
//---------------------------------------------------------------------------
void DrawBulletCnt()
{
	char msg[40];
	bprintf(msg, "%3d", BulletCnt);
	BGDrawStr(BG2Map, 1, 19, (u8*)msg);
}
